/***
 * OW2 FraSCAti Tinfi
 * Copyright (C) 2007-2011 INRIA, USTL
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: frascati@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.ow2.frascati.tinfi.control.content;

import java.lang.annotation.Annotation;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.oasisopen.sca.RequestContext;
import org.oasisopen.sca.ServiceReference;
import org.oasisopen.sca.annotation.Property;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.Interface;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.api.control.LifeCycleController;
import org.objectweb.fractal.api.factory.InstantiationException;
import org.objectweb.fractal.api.type.ComponentType;
import org.objectweb.fractal.api.type.InterfaceType;
import org.objectweb.fractal.julia.Controller;
import org.objectweb.fractal.julia.InitializationContext;
import org.objectweb.fractal.julia.factory.ChainedInstantiationException;
import org.ow2.frascati.tinfi.TinfiRuntimeException;
import org.ow2.frascati.tinfi.api.control.ContentInstantiationException;
import org.ow2.frascati.tinfi.control.content.scope.CompositeScopeManager;
import org.ow2.frascati.tinfi.control.content.scope.ScopeManager;
import org.ow2.frascati.tinfi.control.content.scope.StatelessScopeManager;
import org.ow2.frascati.tinfi.oasis.ServiceReferenceImpl;
import org.ow2.frascati.tinfi.osoa.ConversationScopeManager;
import org.ow2.frascati.tinfi.osoa.RequestScopeManager;
import org.ow2.frascati.tinfi.reflect.InjectionPoint;
import org.ow2.frascati.tinfi.reflect.InjectionPointMap;

/**
 * Mixin layer for implementing the {@link SCAExtendedContentController}
 * interface.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 */
public abstract class SCAContentControllerMixin
implements Controller, SCAExtendedContentController {
    
    /**
     * The metadata for the content class associated to the component controlled
     * by the current controller.
     */
    private ContentClassMetaData ccmd;
    
    /**
     * The scope manager for the component controlled by the current controller.
     */
    private ScopeManager sm;

    /**
     * The properties declared by the content class associated with this
     * controller. The index is the property name and the value is the property
     * type.
     */
    private Map<String,Class<?>> props = new HashMap<String,Class<?>>();

    
    // -------------------------------------------------------------------------
    // Implementation of the Controller interface
    // -------------------------------------------------------------------------

    public void initFcController( final InitializationContext ic )
    throws InstantiationException {
        
        /*
         * Propagate the invocation to other mixin layers.
         */
        _super_initFcController(ic);
        
        /*
         * Initialize this layer.
         * Do it after the propagation in order for _this_XXX fields to be
         * initialized.
         */
        Class<?> c = (Class<?>) ic.content;
        try {
            innerSetFcContentClass(c);
        }
        catch (IllegalContentClassMetaData e) {
            throw new ChainedInstantiationException(e,null,e.getMessage());
        }
    }
  
    
    // -------------------------------------------------------------------------
    // Implementation of the SCAContentController interface
    // -------------------------------------------------------------------------
    
    public void setFcContentClass( Class<?> c )
    throws IllegalLifeCycleException, ContentInstantiationException {
        
        /*
         * Lifecycle check.
         */
        String state = _this_weaveableOptLC.getFcState();
        if( ! state.equals(LifeCycleController.STOPPED) ) {
            String msg = "Component should be stopped";
            throw new IllegalLifeCycleException(msg);
        }
        
        /*
         * Content class meta data and scope manager retrievals.
         */
        try {
            innerSetFcContentClass(c);
        }
        catch (IllegalContentClassMetaData e) {
            throw new ContentInstantiationException(e);
        }
    }

    public Class<?> getFcContentClass() {
    	return ccmd.fcContentClass;
    }
    
    public Object getFcContent() throws ContentInstantiationException {
        return sm.getFcContent();
    }
    
    
    // -------------------------------------------------------------------------
    // Implementation of the SCAExtendedContentController interface
    // -------------------------------------------------------------------------
    
    public void releaseFcContent( Object content, boolean isEndMethod ) {
        sm.releaseFcContent(content,isEndMethod);
    }
    
    public void pushRequestContext( RequestContext rc ) {
        sm.pushRequestContext(rc);
    }
    
    public void popRequestContext() {
        sm.popRequestContext();
    }
    
    public RequestContext peekRequestContext() {
        return sm.peekRequestContext();
    }
    
    public void eagerInit() throws ContentInstantiationException {
        
    	if( ! ccmd.eagerinit ) {
        	return;
        }
        
        if( !(sm instanceof CompositeScopeManager) ) {
            String msg =
                "Eager initialization is meaningless in scopes other than "+
                "COMPOSITE (scope is: "+ccmd.scope+")";
            throw new ContentInstantiationException(msg);
        }
        
        /*
         * Retrieve the content to trigger its initialization.
         */
        sm.getFcContent();
    }
    
    public void start() throws ContentInstantiationException {
        
    	if( ccmd.startMethod == null ) {
        	return;
        }
        
        Object[] contents = sm.getFcCurrentContents();
        for (Object content : contents) {
			try {
				ccmd.startMethod.invoke(content);
			}
			catch (IllegalAccessException e) {
				throw new ContentInstantiationException(e);
			}
			catch (InvocationTargetException e) {
				throw new ContentInstantiationException(e);
			}
		}
    }

    public void stop() throws ContentInstantiationException {
        
    	if( ccmd.stopMethod == null ) {
        	return;
        }
        
        Object[] contents = sm.getFcCurrentContents();
        for (Object content : contents) {
			try {
				ccmd.stopMethod.invoke(content);
			}
			catch (IllegalAccessException e) {
				throw new ContentInstantiationException(e);
			}
			catch (InvocationTargetException e) {
				throw new ContentInstantiationException(e);
			}
		}
    }

    public boolean containsPropertyName( String name ) {
        boolean b = props.containsKey(name);
        return b;
    }
    
    public String[] getPropertyNames() {
    	Set<String> propnames = props.keySet();
        String[] names = propnames.toArray( new String[propnames.size()] );
        return names;
    }
    
    public Class<?> getPropertyType( String name ) {
    	Class<?> type = props.get(name);
    	return type;
    }

    public void setPropertyValue( String name, Object value ) {
    	
    	InjectionPoint ip = ccmd.props.get(name);
    	if( ip == null ) {
    		// No such property. Do nothing.
    		return;
    	}
    	
    	Object[] contents = sm.getFcCurrentContents();    	
    	for (Object content : contents) {
			set(ip,content,value);
		}
    }
    
    public void setReferenceValue( String name, ServiceReference<?> value ) {

    	InjectionPoint ip = getInjectionPoint(name);
    	if( ip == null ) {
    		/*
    		 * No injection point. This may be the case for references which are
    		 * injected with the @Constructor annotation. See for example
    		 * ConstructorClientImpl. Do nothing.
    		 */
    		return;
    	}
    	
    	Class<?> iptype = ip.getType();
    	Object[] contents = sm.getFcCurrentContents();    	
    	for (Object content : contents) {
    		
    		if( List.class.isAssignableFrom(iptype) ) {
    			
    			// SCA style collection reference
    			
    			List<ServiceReference<?>> srs =
    				(List<ServiceReference<?>>) get(ip,content);
    			
    			/*
    			 * If a service reference has already been recorded for the same
    			 * name, we don't need to add it again. The binding may have
    			 * changed but in this case, this is the delegate field of the
    			 * component output interceptor which has been updated. The
    			 * service reference object which is injected in the injection
    			 * point does not need to be changed. The chain service
    			 * reference > component interface > component interceptor is
    			 * still valid.
    			 */
    			for (ServiceReference<?> sr : srs) {
    				ServiceReferenceImpl<?> sri = (ServiceReferenceImpl<?>)sr;
                    Object o = sri._getDelegate();
					Interface itf = (Interface) o;
					String itfname = itf.getFcItfName();
					if( itfname.equals(name) ) {
						return;
					}
				}
    			
    			srs.add(value);
    		}
    		else if( Map.class.isAssignableFrom(iptype) ) {
    			
    			// Fraclet style collection reference
    			
    			Map<String,ServiceReference<?>> srs =
    				(Map<String,ServiceReference<?>>) get(ip,content);
    			srs.put(name,value);
    		}
    		else {
    			// Singleton reference
    			set(ip,content,value);
    		}
		}
    }

    public void unsetReferenceValue( String name ) {
    	
    	InjectionPoint ip = getInjectionPoint(name);
    	if( ip == null ) {
    		/*
    		 * No injection point. This may be the case for references which are
    		 * injected with the @Constructor annotation. See for example
    		 * ConstructorClientImpl. Do nothing.
    		 */
    		return;
    	}
    	
    	Class<?> iptype = ip.getType();    	
    	Object[] contents = sm.getFcCurrentContents();    	
    	for (Object content : contents) {
    		
    		if( List.class.isAssignableFrom(iptype) ) {
    			
    			// SCA style collection reference
    			
    			List<ServiceReference<?>> srs =
    				(List<ServiceReference<?>>) get(ip,content);
    			
    			ServiceReference<?> toBeRemoved = null;
    			for (ServiceReference<?> sr : srs) {
    				ServiceReferenceImpl<?> sri = (ServiceReferenceImpl<?>)sr;
                    Object o = sri._getDelegate();
					Interface itf = (Interface) o;
					String itfname = itf.getFcItfName();
					if( itfname.equals(name) ) {
						toBeRemoved = sr;
					}
				}
    			
    			srs.remove(toBeRemoved);
    		}
    		else if( Map.class.isAssignableFrom(iptype) ) {
    			
    			// Fraclet style collection reference
    			
    			Map<String,ServiceReference<?>> srs =
    				(Map<String,ServiceReference<?>>) get(ip,content);
    			srs.remove(name);
    		}
    		else {
    			// Singleton reference
    			set(ip,content,null);
    		}
		}
    }
    
    
    // -------------------------------------------------------------------------
    // Implementation specific
    // -------------------------------------------------------------------------

    /**
     * Declare the content class which should be used.
     * 
     * This operation can only be performed if the component is stopped.
     * This method is synchronized to prevent inconsistent concurrent changes.
     * 
     * @since 1.0
     */
    private synchronized void innerSetFcContentClass( Class<?> c )
    throws IllegalContentClassMetaData {
        
        /*
         * Content class meta data and scope manager retrievals.
         */
        ccmd = ContentClassMetaData.get(c);
        
        String scope = ccmd.scope==null ? null : ccmd.scope.toUpperCase();
        if( scope == null || scope.equals("STATELESS") ) {
            sm = new StatelessScopeManager(_this_weaveableC,ccmd);
        }
        else if( scope.equals("REQUEST") ) {
            sm = new RequestScopeManager(_this_weaveableC,ccmd);
        }
        else if( scope.equals("COMPOSITE") ) {
            sm = new CompositeScopeManager(_this_weaveableC,ccmd);
        }
        else if( scope.equals("CONVERSATION") ) {
            sm = new ConversationScopeManager(_this_weaveableC,ccmd);
        }
        else {
            String msg = "Unsupported scope: "+ccmd.scope;
            throw new IllegalContentClassMetaData(msg);
        }
        
        /*
         * Initialize property names and types.
         * 
         * Firstly, search in the @Property annotated elements (fields and
         * setters.)
         */
        props = new HashMap<String,Class<?>>();
        InjectionPointMap ipm = ccmd.props;
        Set<String> ipnames = ipm.keySet();
        for (String ipname : ipnames) {
			InjectionPoint ip = ipm.get(ipname);
			Class<?> type = ip.getType();
			props.put(ipname,type);
		}
        
        /*
         * Secondly, search in the @Property annotated parameters of the
         * @Constructor annotated constructor.
         */
        Constructor<?> ctr = ccmd.constructorAnnotatedElement;
        if( ctr != null ) {
                        
            Class<?>[] ptypes = ctr.getParameterTypes();
            Annotation[][] psannots = ctr.getParameterAnnotations();            
            for (int i = 0; i < ptypes.length; i++) {
            	// By definition, ptypes.length == psannots.length
				Class<?> ptype = ptypes[i];
				Annotation[] pannots = psannots[i];
                for (Annotation pannot : pannots) {
                    if( pannot instanceof Property ) {
                        Property prop = (Property) pannot;
                        String propname = prop.name();
                        props.put(propname,ptype);
                    }
                }
			}
        }
    }
    
    private InjectionPoint getInjectionPoint( String name ) {
    	/*
    	 * Do not mimic setPropertyValue which iterates on ccmd.props and
    	 * iterate on ccmd.refs. The reason is that one may find the case where
    	 * the component defines e.g. an opt singleton interface and an opts
    	 * collection interfaces. When searching e.g. for opts0 both injection
    	 * point will match and one may ends up with opt instead of opts. 
    	 */
    	
    	ComponentType ct = (ComponentType) _this_weaveableC.getFcType();
    	InterfaceType it = null;
    	try {
			it = ct.getFcInterfaceType(name);
		}
    	catch (NoSuchInterfaceException e) {
    		// No such interface, neither singleton, nor collection.
    		return null;
		}
    	String itname = it.getFcItfName();
    	
    	/*
    	 * For singleton interfaces, itname equals name.
    	 * For collection interfaces, itname equals the prefix of the collection
    	 * for name.
    	 */
    	InjectionPoint ip = ccmd.refs.get(itname);
    	return ip;
    }
    
    /**
     * Wraps exception thrown by {@link InjectionPoint#get(Object)} into {@link
     * TinfiRuntimeException}s.
     * 
     *  @since 1.3.1
     */
    private static Object get( InjectionPoint ip, Object content )
    throws TinfiRuntimeException {
    	
		try {
			return ip.get(content);
		}
		catch (IllegalAccessException e) {
			throw new TinfiRuntimeException(e);
		}
		catch (InvocationTargetException e) {
			throw new TinfiRuntimeException(e);
		}
    }
    
    /**
     * Wraps exception thrown by {@link InjectionPoint#set(Object, Object)} into
     * {@link TinfiRuntimeException}s.
     * 
     *  @since 1.3.1
     */
    private static void set( InjectionPoint ip, Object content, Object value )
    throws TinfiRuntimeException {
    	
		try {
			ip.set(content,value);
		}
		catch (IllegalAccessException e) {
			throw new TinfiRuntimeException(e);
		}
		catch (InvocationTargetException e) {
			throw new TinfiRuntimeException(e);
		}
    }
    

    // -------------------------------------------------------------------------
    // Fields and methods required by the mixin class in the base class
    // -------------------------------------------------------------------------

    public abstract void _super_initFcController(InitializationContext ic)
    throws InstantiationException;
    
    public Component _this_weaveableC;
    public LifeCycleController _this_weaveableOptLC;
}
